# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from pgpy import PGPKey, PGPMessage
from pgpy.errors import PGPDecryptionError
import warnings

from config_yourself.provider.datakey import DataKeyService
from config_yourself.exceptions import InvalidConfig, BlinkConfigException


class Service(DataKeyService):
    """A GPG CryptoService

    Provide the ``privateKey`` secret to :py:class:`~config_yourself.Config` when loading to initialize this provider. It should contain the bytes of a GPG private key. Provide the ``privateKeyPassword`` if the key is password protected.
    """

    def __init__(self, config, secrets):
        try:
            dataKeyBytes = config.get("key", None)
        except AttributeError as e:
            raise InvalidConfig(original=e)

        if not dataKeyBytes or dataKeyBytes == "":
            msg = (
                "crypto.key is not a string, "
                "Remove the crypto property if encryption is not needed"
            )
            raise InvalidConfig(message=msg)

        privateKey = secrets.get("privateKey", None)
        if not privateKey or privateKey == "":
            msg = (
                "No gpg private key provided for decryption. "
                "Remove the crypto property if encryption is not needed"
            )
            raise InvalidConfig(message=msg)

        gpgKey = PGPKey()
        gpgKey.parse(privateKey)

        password = secrets.get("privateKeyPassword", None)
        if password:
            try:
                gpgKey.unlock(password)
            except PGPDecryptionError as err:
                raise BadGPGKeyPasswordError(gpgKey.userids[0])

        with warnings.catch_warnings():
            # prevents warning of type `UserWarning: Message was encrypted with this key's subkey: ...`
            warnings.simplefilter("ignore", category=UserWarning)
            dataKey = gpgKey.decrypt(PGPMessage.from_blob(dataKeyBytes)).message

        DataKeyService.__init__(self, dataKey)


class BadGPGKeyPasswordError(BlinkConfigException):
    def __init__(self, key):

        msg = "Invalid password for key <{}>".format(key)

        self.message = msg
        super(BadGPGKeyPasswordError, self).__init__(msg)



# coding=utf-8
from __future__ import print_function
from base64 import b64decode
from tests.conftest import KEYS, SECRETS

import pytest
from config_yourself.provider.gpg import Service


class TestGPG:
    @pytest.mark.parametrize(
        "config,secrets,errors",
        [
            ({}, None, True),
            ("nope", None, True),
            ({"key": "badkey"}, None, True),
            ({"key": KEYS["gpg"]}, SECRETS["gpg"], False),
        ],
    )
    def test_constuctor(self, config, secrets, errors):
        if errors:
            with pytest.raises(Exception):
                Service(config, secrets)
        else:
            s = Service(config, secrets)
            assert callable(s.decrypt)

    def test_decrypt(self):
        encrypted = (
            "on//Ow0gJ3gRBzKO7Q0jq/75v6oIlUKZV3lJOuE3CYfEj/hHlYdX5O3WF10dseH1y8QIgi2n"
        )
        s = Service({"key": KEYS["gpg"]}, SECRETS["gpg"])
        plainText = s.decrypt(b64decode(encrypted))
        assert plainText == b"secret"



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass


class CryptoService(with_metaclass(ABCMeta, object)):
    """The CryptoService abstract class must be implemented by providers
    """

    @abstractmethod
    def decrypt(self, ciphertext):  # pragma: no cover
        """The decrypt method takes a ciphertext string and returns the plaintext contents of it"""
        pass



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from config_yourself.Config import Config, AppConfig
import config_yourself.load as load
import config_yourself.exceptions as exceptions



# coding=utf-8
from __future__ import print_function
import pytest
from moto import mock_kms
from config_yourself import load


# noinspection PyClassHasNoInit
@mock_kms
class TestLoaders(object):
    @pytest.mark.parametrize(
        ["env", "result"],
        [
            [{"BLINKCONFIG.string": "value"}, {"string": "value"}],
            [{"BLINKCONFIG.boolean": "true"}, {"boolean": True}],
            [{"BLINKCONFIG.nested.boolean": "true"}, {"nested": {"boolean": True}}],
        ],
    )
    def test_env_loader(self, monkeypatch, env, result):
        for k, v in env.items():
            monkeypatch.setenv(k, v)
        assert load.env() == result



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

import re
import boto3
import botocore
from config_yourself.exceptions import InvalidConfig, BlinkConfigException
from .base import CryptoService


class Service(CryptoService):
    """The original KMS CryptoService

    This provider does not need ``secrets``, and will use the AWS default credential chain to get credentials.
    """

    def __init__(self, config, _secrets):
        # secrets are ignored by the kms provider
        self._client = None
        self._region = None
        self._key = None
        self._setup_client(config)

    def _setup_client(self, crypto):
        # Setup the region for kms by looking at `crypto.region` and fallback on parsing
        # the key ARN and extracting the region from it
        try:
            region = crypto.get("region", None)
        except AttributeError as e:
            # When the crypto property exists, but is not a dict, we raise this error.
            raise InvalidConfig(original=e)

        try:
            if not region and "key" in crypto and ":" in crypto["key"]:
                # try and fetch the region name from the key's arn if specified
                # arn:aws:kms:{region}:{account}:{key_name}
                arn = crypto["key"].split(":")
                if len(arn) == 6:
                    region = crypto["key"].split(":")[3]
                else:
                    raise IndexError
        except TypeError:
            # crypto.key is not a string, which is likely a pebkac error!
            msg = (
                "crypto.key is not a string, "
                "please remove the crypto property if no encryption is needed"
            )
            raise InvalidConfig(message=msg)
        except IndexError:
            # crypto.key is likely in the non fq ARN alias/form
            msg = (
                "Could not parse region name from crypto.key. Please ensure you have "
                "specified a fully-qualifed KSM key ARN to replace `{}` before continuing".format(
                    crypto["key"]
                )
            )
            raise InvalidConfig(message=msg)

        if not region or region == "":
            # region not set or empty
            msg = (
                "Unable to resolve AWS region to query when decrypting. "
                "Please use specify a full KMS key ARN in `crypto.key` or set a specific region "
                "by adding a `crypto.region`"
            )
            raise InvalidConfig(message=msg)

        self._region = region
        self._key = crypto["key"]
        self._client = boto3.client("kms", region_name=self._region)

    def decrypt(self, ciphertext):
        try:
            res = self._client.decrypt(CiphertextBlob=ciphertext)
            return res["Plaintext"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "AccessDeniedException":
                # when we get this error, the most likely case is region mismatch or bonafide
                # permission error, let's make this error helpful by showing which credentials
                # we tried to use during decryption
                try:
                    creds = boto3.client(
                        "sts", region_name=self._region
                    ).get_caller_identity()["Arn"]
                except Exception:
                    creds = None

                raise KMSDecryptError(
                    str(self._region), self._key, creds=creds, original=e
                )
            else:
                raise


class KMSDecryptError(BlinkConfigException):
    def __init__(self, region, key, creds=None, original=None):
        msg = ""
        if original:
            msg = "{}. ".format(original)

        msg += "Failed to decrypt in region <{}> with key <{}>".format(region, key)

        if creds:
            msg += " using credentials <{}>".format(creds)
        else:
            msg += " with no known credentials"

        self.message = msg
        super(KMSDecryptError, self).__init__(msg)



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from config_yourself.exceptions import InvalidConfig, BlinkConfigException
from config_yourself.provider.datakey import (
    DataKeyService,
    DataKeyDecryptError,
    decrypt_key_with_password,
)
from base64 import b64decode
from binascii import Error


class Service(DataKeyService):
    """A Password CryptoService

    Provide the ``password`` secret to :py:class:`~config_yourself.Config` when loading to initialize this provider.
    """

    def __init__(self, config, secrets):
        try:
            key = config.get("key", None)
        except AttributeError as e:
            raise InvalidConfig(original=e)

        if not key or key == "":
            msg = (
                "crypto.key is empty, "
                "remove the crypto property if no encryption is needed"
            )
            raise InvalidConfig(message=msg)

        try:
            keyBytes = b64decode(key)
        except Error:
            raise InvalidConfig(message="Could not read file key as base64")

        password = secrets.get("password", None)
        if not password or password == "":
            msg = (
                "No password provided for decryption. "
                "Remove the crypto property if no encryption is needed"
            )
            raise InvalidConfig(message=msg)

        try:
            # decrypt the file's key with the supplied password
            dataKey = decrypt_key_with_password(keyBytes, password)
        except DataKeyDecryptError:
            raise PasswordDecryptError(key)
        DataKeyService.__init__(self, dataKey)


class PasswordDecryptError(BlinkConfigException):
    def __init__(self, key):

        msg = "Failed to decrypt with key <{}> and a supplied password".format(key)

        self.message = msg
        super(PasswordDecryptError, self).__init__(msg)



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

try:  # pragma: no cover
    # YAY PYTHON!
    # noinspection PyPackageRequirements
    from collections.abc import Mapping, Hashable
except ImportError:  # pragma: no cover
    try:
        # noinspection PyPackageRequirements
        from collections import Mapping, Hashable
    except ImportError:
        # noinspection PyPackageRequirements
        from future.moves.collections import Mapping, Hashable
import warnings


def merge_dicts(src, dest, parent=[]):
    for key in dest:
        full_path = parent + [key]
        if key in src:
            if isinstance(src[key], Mapping) and isinstance(dest[key], Mapping):
                if src[key].get("encrypted", False):
                    msg = (
                        "Overriding an encrypted value at <{}>. This can lead to trouble, since these "
                        "values were likely encrypted with different keys!"
                    )
                    warnings.warn(msg.format(".".join(full_path)), DeprecationWarning)
                src[key] = merge_dicts(src[key], dest[key], parent=full_path)
            elif src[key] == dest[key]:
                pass  # same leaf value
            else:
                src[key] = dest[key]
        else:
            # don't let overrides create keys
            msg = (
                "Trying to override a value on a non-existing default! "
                "The value at path <{}> will be ignored".format(".".join(full_path))
            )
            warnings.warn(msg, DeprecationWarning)
            pass
    return src



# coding=utf-8
from __future__ import print_function

import pytest

# mock_kms just base64 encodes values
from moto import mock_kms
from copy import deepcopy
import warnings
from tests.conftest import add_kms_entries, get_config_yourself, KEYS, SECRETS
import config_yourself.exceptions as errors

# noinspection PyClassHasNoInit
@mock_kms
class TestBlinkConfig(object):
    @pytest.mark.parametrize(
        "input_config",
        [
            {"key": "value"},
            {"key": {"encrypted": True, "ciphertext": "Wlc1amNubHdkRzFs"}},
            {"key": {"nested": {"value": True}}},
            {"key": [{"nested": True}, {"nested": False}]},
        ],
    )
    def test_verify_returned_encrypted_config_can_not_modify_original(
        self, input_config
    ):
        input_config_with_kms = add_kms_entries(input_config)
        original = input_config_with_kms["key"]
        with warnings.catch_warnings(record=True):
            config_yourself = get_config_yourself(
                input_config_with_kms, {"key": "overriden"}
            )
        with pytest.raises(TypeError):
            config_yourself["foo"] = "bar"

        assert input_config_with_kms["key"] == original
        assert config_yourself["key"] == "overriden"

    @pytest.mark.parametrize(
        "source,override",
        [
            (
                {"value": True, "key": {"nested": {"value": True, "unchanged": True}}},
                {"value": False, "key": {"nested": {"value": False}}},
            )
        ],
    )
    def test_verify_overrides_are_not_destructive(self, source, override):
        original_source = deepcopy(source)
        with warnings.catch_warnings(record=True):
            resulting_config = get_config_yourself(source, override)

        # the original object was not modified
        assert source == original_source
        # the resulting config has been overridden
        assert not resulting_config["key"]["nested"]["value"]
        # make sure the nested value that is defined in source and not defined in the override is present
        assert "unchanged" in resulting_config["key"]["nested"]
        assert resulting_config["key"]["nested"]["unchanged"]

    def test_overrides_warn_on_potential_errors(self):
        with warnings.catch_warnings(record=True) as warns:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            crypto = {"key": "<key>"}
            source = {
                "crypto": crypto,
                "key": {"encrypted": True, "ciphertext": "Wlc1amNubHdkRzFs"},
            }
            override = {
                "crypto": crypto,
                "key": {"encrypted": True, "ciphertext": "Y3JhcCEK"},
                "ignored": True,
            }

            with pytest.raises(errors.InvalidConfig):
                cfg = get_config_yourself(source, override)

            all_warnings = "\n".join([str(w.message) for w in warns])
            assert "<key>" in all_warnings
            assert "<ignored>" in all_warnings
            assert "defaulting to `kms`" in all_warnings

    def test_decrypts(self):
        secret = {
            "ciphertext": "Ds84SJ3l3l0Xz2CSlo3Ap0MF3bNTFA/+IVZ8uLIk7YhEC/mVrV5je6LlaBst7Xj2ggWs9pN6",
            "encrypted": "true",
        }
        cfg = get_config_yourself(
            {
                "crypto": {
                    "provider": "password",
                    "key": "dRVeKIGwNwJU0LQ69eUwsUyQRCyPh7DMRiU1bK//LtoMg81SHohquBR9S5fSYa6Uz+yvAHrx2KjBzS+0QXs6bM6fDJVpNodXOgA5XtNMV+iA8hVZlkC12cnfsNw=",
                },
                "secret": secret,
                "nested": {"secret": secret},
                "list": [{"secret": secret}, secret],
            },
            password="password",
        )

        assert cfg["secret"] == "secret"
        assert cfg["nested"]["secret"] == "secret"
        assert cfg["list"][0]["secret"] == "secret"
        assert cfg["list"][1] == "secret"
        assert len(cfg) == 4

    def test_fails_on_bad_secret(self):
        with pytest.raises(errors.DecryptError) as err:
            get_config_yourself(
                {
                    "crypto": {"provider": "password", "key": KEYS["password"]},
                    "secret": {"ciphertext": "bad-secret", "encrypted": "true"},
                },
                password="password",
            )

    def test_fails_on_unknown_provider(self):
        with pytest.raises(ImportError) as err:
            get_config_yourself(
                {"crypto": {"provider": "fake-provider", "key": "adsf"}}
            )



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from base64 import b64decode
from os import environ, path

import json

try:  # pragma: no cover
    # YAY PYTHON!
    # noinspection PyPackageRequirements
    from collections.abc import Mapping, Hashable
except ImportError:  # pragma: no cover
    try:
        # noinspection PyPackageRequirements
        from collections import Mapping, Hashable
    except ImportError:
        # noinspection PyPackageRequirements
        from future.moves.collections import Mapping, Hashable
from functools import reduce

from six import iteritems

import config_yourself.exceptions as exceptions
import config_yourself.provider as provider
import config_yourself.load as load
from config_yourself.util import merge_dicts


class Config(Mapping, Hashable):
    """A Config object

    The configs supplied are merged sequentially, the resulting ``dict`` is decrypted recursively and
    frozen. ``Config`` objects are proxies for that resulting ``dict``, and behave just like it

    :param dict[str,Any] default_config: A dictionary with default values
    :param List[dict[str,Any]] configs: Any number of overrides to apply over `default_config`
    :param str password: The password string when `crypto.provider` is password
    :param str privateKey: The armored contents of the private GPG key when `crypto.provider` is gpg

    :return: A decrypted config file as a frozen dictionary
    """

    def __init__(self, default_config, *configs, **secrets):
        self._hash = None
        self._provider = None

        cfg = self._merge_configs(default_config, configs)
        crypto = cfg.get("crypto", None)
        if not crypto:
            # no crypto means no secrets
            self._data = cfg
            return

        self._provider = provider.Load(crypto, secrets)
        self._data = self._decrypt(cfg)

    def _merge_configs(self, default_config, overrides):
        if len(overrides) > 0:
            # copy the first element to prevent modifying it. Deepcopy is way slower for large dicts
            # such as backend configs, and both `copy` and `OrderedDict` keep nested references
            # this hack is fast, but a hack nonetheless
            copies = tuple((json.loads(json.dumps(default_config)),) + overrides)
            return reduce(merge_dicts, copies)
        else:
            # when we're not using overrides, we don't care about copying, since the decryption
            # process copies on iteration
            return default_config

    def _decrypt(self, node, parents=[]):
        if isinstance(node, Mapping):
            if "encrypted" in node:
                try:
                    cipherBytes = b64decode(node["ciphertext"])
                    return self._provider.decrypt(cipherBytes).decode("utf-8")
                except Exception as e:
                    raise exceptions.DecryptError(parents, e)

            decrypted_dict = dict()
            for key, inner_value in iteritems(node):
                full_path = parents + [key]
                decrypted_dict[key] = self._decrypt(inner_value, parents=full_path)
            return decrypted_dict

        elif isinstance(node, list):
            decrypted_list = []
            for index, subNode in enumerate(node):
                full_path = parents + [str(index)]
                decrypted_list.append(self._decrypt(subNode, parents=full_path))
            return decrypted_list

        else:
            return node

    # basically make this class into a frozen dict
    # taken from http://stackoverflow.com/a/2704866

    def __iter__(self):  # pragma: no cover
        return iter(self._data)

    def __len__(self):
        """Get the size of this dict"""
        return len(self._data)

    def __getitem__(self, key):
        """Get the value for a top level config item"""
        return self._data[key]

    def __hash__(self):  # pragma: no cover
        if self._hash is None:
            self._hash = 0
            # noinspection PyTypeChecker
            for pair in iteritems(self):
                self._hash ^= hash(pair)
        return self._hash


def AppConfig(config_file=None, before=None, after=None, config_folder="./config"):
    """Load a sequence of resolved config file names

    When using config_yourself in Flask or Django, this helper will load:

    * the ``"{config_folder}/defaults.yml"`` file,
    * a file located at ``os.environ.get("CONFIG_FILE", path.join(config_folder, "local.yml"))``
    * a ``"{config_folder}/personal.yml"`` file, if it exists


    All parameters are optional, and allow tweaking of the locations and extensions used for config files.

    :param Optional[str] config_file: the path to the main config file to load, usually. (default: ``os.environ.get("CONFIG_FILE", "./config/local.yml")``)
    :param Optional[list[str]] before: File names to load before ``config_file`` (default: `['default']`)
    :param Optional[list[str]] after: File names to apply after ``config_file`` if they exist (default: `['personal']`)
    :param Optional[str] config_folder: Base folder path to look for files in, (default: `./config`)

    :returns: A Config object
    :rtype: :py:class:`~config_yourself.Config`
    """
    if config_file is None:
        config_file = environ.get("CONFIG_FILE", path.join(config_folder, "local.yml"))

    if before is None:
        before = ["defaults"]

    if after is None:
        after = ["personal"]

    # the chain of files to merge before decrypting
    chain = []
    # extract some details about the main file
    config_file = path.abspath(config_file)
    if config_folder is None:
        config_folder = path.dirname(config_file)
    _main_ext = path.splitext(config_file)[1]
    name_to_path = lambda name: path.join(
        config_folder, "{}.{}".format(name, _main_ext)
    )

    # Add the resolved before files
    chain = [name_to_path(file_name) for file_name in before]
    # Add the main config file
    chain.append(config_file)

    # Add after files if they exist
    for f in [name_to_path(file_name) for file_name in after]:
        if path.exists(f):
            chain.append(f)

    return Config(*[load.file(file_path) for file_path in chain])



# coding=utf-8
from __future__ import print_function
from base64 import b64decode
from tests.conftest import KEYS, SECRETS

import pytest
from config_yourself.provider.password import Service
from tests.conftest import KEYS, SECRETS


class TestPassword:
    @pytest.mark.parametrize(
        "config,secrets,errors",
        [
            ({}, None, True),
            ("nope", None, True),
            ({"key": "badkey"}, None, True),
            ({"key": KEYS["password"]}, {"password": "badpassword"}, True),
            ({"key": KEYS["password"]}, {"no-password": ""}, True),
            ({"key": KEYS["password"]}, SECRETS["password"], False),
        ],
    )
    def test_constuctor(self, config, secrets, errors):
        if errors:
            with pytest.raises(Exception):
                Service(config, secrets)
        else:
            s = Service(config, secrets)
            assert callable(s.decrypt)

    def test_decrypt(self):
        encrypted = (
            "Ds84SJ3l3l0Xz2CSlo3Ap0MF3bNTFA/+IVZ8uLIk7YhEC/mVrV5je6LlaBst7Xj2ggWs9pN6"
        )
        s = Service({"key": KEYS["password"]}, SECRETS["password"])
        plainText = s.decrypt(b64decode(encrypted))
        assert plainText == b"secret"



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

__all__ = ["file", "env"]

import json
import os
from yaml import safe_load


def env(prefix="BLINKCONFIG"):
    """Loads all environment variables prefixed with `prefix_`

    :param str prefix: The prefix to use when filtering environment variables. (default: `BLINKCONFIG`)

    :return: A dictionary of loaded values from the environment
    :rtype: dict[str,Any]
    """

    prefix = prefix + "."
    keys = [k.replace(prefix, "", 1) for k in os.environ if k.startswith(prefix)]

    data = dict()
    for key in keys:
        path = key.split(".")
        node = data

        for idx, leaf in enumerate(path):
            if idx + 1 == len(path):
                value = os.environ[prefix + key]
                try:
                    value = json.loads(value)
                except ValueError:
                    pass

                node[leaf] = value
            else:
                node[leaf] = dict()
                node = node[leaf]

    return data


def file(path):
    """returns the dict representation of yaml file at `path`

    :param str path: The path of the file to parse as YAML.

    :return: A dictionary of loaded values from the file
    :rtype: dict[str,Any]
    """

    with open(path, "r") as config_file:
        return safe_load(config_file.read())



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from importlib import import_module
import warnings
from future.utils import raise_from


def Load(crypto, secrets=None):
    """Load a crypto provider, initializing it with secrets

    :param dict[str,Union[str,List]] crypto: The Config's `crypto` node
    :param dict[str,str] secrets: The runtime provided secrets

    :return: An initialized CryptoService
    :rtype: config_yourself.provider.CryptoService
    """
    providerName = crypto.get("provider", None)
    if not providerName:
        msg = "crypto.provider undefined in config file, defaulting to `kms`"
        warnings.warn(msg, DeprecationWarning)
        providerName = "kms"

    module_path = "config_yourself.provider.{}".format(providerName)
    module = import_module(module_path)

    try:
        provider = getattr(module, "Service")
    except AttributeError as err:  # pragma: no cover
        msg = 'Provider "%s" does not define a Service class' % (module_path)
        raise_from(ImportError(msg), err)

    return provider(crypto, secrets)



class BlinkConfigException(Exception):
    """The base BlinkConfig Exception"""

    pass


class DecryptError(BlinkConfigException):
    """A value could not be decrypted"""

    def __init__(self, full_path, original=None):
        msg = "Could not decrypt path {}".format(".".join(full_path))
        super(DecryptError, self).__init__(msg)


class InvalidConfig(BlinkConfigException):
    """A supplied config file was not valid"""

    def __init__(
        self,
        original=None,
        message="crypto property is not an object of key and region.",
    ):
        if original:
            message = "{}\n{}".format(message, original)
        self.message = message
        super(InvalidConfig, self).__init__(message)



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

import os
import subprocess

from sys import version_info
from setuptools import setup, find_packages
from setuptools_scm import get_version

version = get_version()
with open("./version.py", "w") as fh:
    message = "{header}{nl}__version__='{ver}'{nl}".format(
        header="# generated by setup.py", ver=version, nl=os.linesep
    )
    fh.write(message)

deps = {
    # required by datakey
    "aes": "cryptography >=2.8,<3.0",
    # gpg
    "gpg": "pgpy >=0.4.3,<1.0.0",
    # kms
    "boto": "boto3 >=1.4.4,<2.0.0",
}

if version_info[0] == 2:
    # https://github.com/SecurityInnovation/PGPy/issues/217
    six = ["six ==1.10.0"]
else:
    six = []

setup(
    name="config-yourself",
    description="A python library to decrypt config-yourself files",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    url="https://github.com/blinkhealth/config-yourself-python",
    license="Apache License 2.0",
    author="Will Richard",
    author_email="william.richard@blinkhealth.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Topic :: Security",
        "Topic :: Security :: Cryptography",
        "Topic :: Utilities",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: JavaScript",
    ],
    packages=find_packages(),
    extras_require={
        # gpg requires aes to do datakey decryption
        "gpg": six + [deps["gpg"], deps["aes"]],
        "kms": [deps["boto"]],
        "password": [deps["aes"]],
        "all": six + [deps.values()],
    },
    install_requires=["future >=0.16.0,<1.0.0", "PyYAML >=5.1.1"],
)



# coding=utf-8
from __future__ import print_function

import pytest
from moto import mock_kms
from config_yourself.provider.kms import Service
from tests.conftest import KEYS
from tests.conftest import encrypt_test_value
from base64 import b64decode


@mock_kms
class TestKMS:
    @pytest.mark.parametrize(
        "config,region,errors",
        [
            ({}, None, True),
            ("nope", None, True),
            ({"key": "badkey"}, None, True),
            ({"key": 1}, None, True),
            ({"key": "bad-arn:key/000"}, None, True),
            ({"key": KEYS["kms"]}, "mx-central-1", False),
            ({"key": "key/000", "region": "mx-central-1"}, "mx-central-1", False),
        ],
    )
    def test_constuctor(self, config, region, errors):
        if errors:
            with pytest.raises(Exception):
                Service(config, {})
        else:
            s = Service(config, {})
            assert callable(s.decrypt)
            assert region == "mx-central-1"

    def test_decrypt(self):
        encrypted = encrypt_test_value("secret")
        s = Service({"key": KEYS["kms"]}, None)
        plainText = s.decrypt(b64decode(encrypted["ciphertext"]))
        assert plainText == b"secret"



# Copyright 2018 Blink Health LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

from .base import CryptoService
from builtins import bytes
from config_yourself.exceptions import BlinkConfigException
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

NONCE_BYTE_SIZE = 32
SALT_SIZE = 12


class DataKeyService(CryptoService):
    """The DataKeyService implements a CryptoService that uses an in-file data key to decrypt secrets

    :param bytes key: this file's datakey
    """

    def __init__(self, key):
        self._cipher = AESGCM(key)

    def decrypt(self, ciphertext):
        nonce, data = ciphertext[0:NONCE_BYTE_SIZE], ciphertext[NONCE_BYTE_SIZE:]
        try:
            return self._cipher.decrypt(nonce, data, b"")
        except InvalidTag:
            raise DataKeyDecryptError("Could not decrypt")


def decrypt_key_with_password(key, password):
    """Turns a key and password into a scrypt key

    :param bytes key: The key bytes, containing the salt and encrypted key
    :param str password: The password in plaintext to derive the key from

    :returns: The decrypted key contained in ``key``
    :rtype: bytes
    """
    # get the salt and encrypted key
    salt, encryptedKey = key[0:SALT_SIZE], key[SALT_SIZE:]
    # re-create the file's key with the provided password and salt
    derivedKey = Scrypt(
        salt,
        backend=default_backend(),
        # key length for AES
        length=32,
        # The cost parameter
        n=32 * 1024,
        # same values as go-config-yourself
        r=8,
        p=1,
    ).derive(bytes(password, "utf-8"))
    # Return the decrypted key using the derived key
    return DataKeyService(derivedKey).decrypt(encryptedKey)


class DataKeyDecryptError(BlinkConfigException):
    pass



import pytest
import warnings
from tests.conftest import get_config_yourself, add_kms_entries, encrypt_test_value


class TestBlinkConfig(object):
    @pytest.mark.parametrize(
        "input_config,expected_config",
        [
            (
                {"encrypted_key": encrypt_test_value("encryptme")},
                {"encrypted_key": "encryptme"},
            ),
            (
                {"encrypted_number": encrypt_test_value("042")},
                {"encrypted_number": "042"},
            ),
            (
                {
                    "nested_encrypted_dict": {
                        "outer_dict": {
                            "first_value": encrypt_test_value("hidden"),
                            "second_value": encrypt_test_value("092"),
                        }
                    }
                },
                {
                    "nested_encrypted_dict": {
                        "outer_dict": {"first_value": "hidden", "second_value": "092"}
                    }
                },
            ),
        ],
    )
    def test_config(self, input_config, expected_config):
        input_with_kms_entries = add_kms_entries(input_config)
        expected_with_kms_entries = add_kms_entries(expected_config)
        config_yourself = get_config_yourself(input_with_kms_entries)

        assert config_yourself._data == expected_with_kms_entries

    @pytest.mark.parametrize(
        "input_config, other_config, expected_overridden_config",
        [
            (
                {"both_key": "initial_value"},
                {"both_key": "new_value", "other_key": "other_value"},
                {"both_key": "new_value"},
            ),
            (
                {"change_key": "initial_value", "untouched_key": "untouched_value"},
                {"change_key": "other_value", "other_key": "other_value"},
                {"change_key": "other_value", "untouched_key": "untouched_value"},
            ),
            (
                {
                    "outer-key": {
                        "inner-change-key": "initial-value",
                        "inner-untouched": "untouched-value",
                    }
                },
                {
                    "outer-key": {
                        "inner-change-key": "new-value",
                        "inner-other": "other-value",
                    }
                },
                {
                    "outer-key": {
                        "inner-change-key": "new-value",
                        "inner-untouched": "untouched-value",
                    }
                },
            ),
            (
                {"change-secret": encrypt_test_value("initial-secret")},
                {"change-secret": encrypt_test_value("other-secret")},
                {"change-secret": encrypt_test_value("other-secret")},
            ),
            (
                {
                    "outer-key": {
                        "untouched-secret": encrypt_test_value("my-secret"),
                        "touched-secret": encrypt_test_value("initial-secret"),
                    }
                },
                {
                    "outer-key": {
                        "other-key": "other-value",
                        "touched-secret": encrypt_test_value("new-secret"),
                    }
                },
                {
                    "outer-key": {
                        "untouched-secret": encrypt_test_value("my-secret"),
                        "touched-secret": encrypt_test_value("new-secret"),
                    }
                },
            ),
        ],
    )
    def test_override(self, input_config, other_config, expected_overridden_config):
        with warnings.catch_warnings(record=True):
            input_with_kms_entries = get_config_yourself(
                add_kms_entries(input_config), add_kms_entries(other_config)
            )

            expected_with_kms_entries = get_config_yourself(
                add_kms_entries(expected_overridden_config)
            )

            assert input_with_kms_entries._data == expected_with_kms_entries._data
